/*-
 * Copyright (c) 1992 The Regents of the University of California.
 * All rights reserved.
 *
 * This code is derived from software contributed to Berkeley by
 * James da Silva at the University of Maryland at College Park.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/cdefs.h>
__FBSDID("$FreeBSD: release/9.0.0/lib/libcompat/4.3/re_comp.c 205146 2010-03-14 10:18:58Z ed $");

/*
 * Compatibility routines that implement the old re_comp/re_exec interface in
 * terms of the regcomp/regexec interface.  It's possible that some programs
 * rely on dark corners of re_comp/re_exec and won't work with this version,
 * but most programs should be fine.
 */

#if defined(LIBC_SCCS) && !defined(lint)
static char sccsid[] = "@(#)regex.c	5.1 (Berkeley) 3/29/92";
#endif /* LIBC_SCCS and not lint */

#include <regex.h>
#include <stddef.h>
#include <unistd.h>

static regex_t re_regexp;
static int re_gotexp;
static char re_errstr[100];

char *
re_comp(const char *s)
{
	int rc;

	if (s == NULL || *s == '\0') {
		if (!re_gotexp)
			return __DECONST(char *,
			    "no previous regular expression");
		return (NULL);
	}

	if (re_gotexp) {
		regfree(&re_regexp);
		re_gotexp = 0;
	}

	rc = regcomp(&re_regexp, s, REG_EXTENDED);
	if (rc == 0) {
		re_gotexp = 1;
		return (NULL);
	}

	regerror(rc, &re_regexp, re_errstr, sizeof(re_errstr));
	re_errstr[sizeof(re_errstr) - 1] = '\0';
	return (re_errstr);
}

int
re_exec(const char *s)
{
	int rc;

	if (!re_gotexp)
		return (-1);
	rc = regexec(&re_regexp, s, 0, NULL, 0);
	return (rc == 0 ? 1 : 0);
}
